// ************************************************************************** //
//
//  BornAgain: simulate and fit scattering at grazing incidence
//
//! @file      Samples/inc/LayerInterface.h
//! @brief     Defines class LayerInterface.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2015
//! @authors   Scientific Computing Group at MLZ Garching
//! @authors   C. Durniak, M. Ganeva, G. Pospelov, W. Van Herck, J. Wuttke
//
// ************************************************************************** //

#ifndef LAYERINTERFACE_H
#define LAYERINTERFACE_H

#include "ICompositeSample.h"
#include "LayerRoughness.h"

class Layer;

//! @class LayerInterface
//! @ingroup samples_internal
//! @brief Interface between two layers, possibly with roughness.

class BA_CORE_API_ LayerInterface : public ICompositeSample
{
public:
    virtual ~LayerInterface();

    LayerInterface *clone() const;

    //! calls the ISampleVisitor's visit method
    virtual void accept(ISampleVisitor *visitor) const;

    //! Creates smooth interface between two layers
    static LayerInterface* createSmoothInterface(const Layer *p_layer_top,
                                                 const Layer *p_layer_bottom);

    //! Creates rough interface between two layers
    static LayerInterface* createRoughInterface(const Layer *p_layer_top,
                                                const Layer *p_layer_bottom,
                                                const LayerRoughness& roughness);

    //! Sets link to the layer above the interface.
    void setLayerTop(const Layer* p_layer_top);

    //! Sets link to the layer below the interface.
    void setLayerBottom(const Layer* p_layer_bottom);

    //! Sets links to the layers above and below the interface.
    void setLayersTopBottom(
        const Layer* p_layer_top, const Layer* p_layer_bottom)
    { setLayerTop(p_layer_top); setLayerBottom(p_layer_bottom); }

    //! Sets roughness of the interface.
    void setRoughness(const LayerRoughness& roughness);

    //! Returns roughness of the interface.
    const LayerRoughness *getRoughness() const;

    //! Returns top layer.
    const Layer *getLayerTop() const;

    //! Returns bottom layer.
    const Layer *getLayerBottom() const;

protected:
    void print(std::ostream& ostr) const;

private:
    LayerInterface();

    LayerRoughness *m_roughness;   //!< roughness of the interface
    const Layer *m_LayerTop;       //!< pointer to the layer above interface
    const Layer *m_LayerBottom;    //!< pointer to the layer below interface
};

#endif // LAYERINTERFACE_H


